# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2018-2023 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

__ALL__ = [
  'Neo4jConnector',
  'Neo4jDriver',
  'ODBCConnector',
  'SQLODBCDriver',
]

from .neo4j_connector import Neo4jConnector, Neo4jDriver

try:
    from .odbc import ODBCConnector
    from .odbc import SQLODBCDriver
    from .odbc import MongoODBCDriver
    from .odbc import OracleODBCDriver
    from .odbc import SAPODBCDriver
    from .odbc import SnowflakeODBCDriver
except ImportError:
    pass



# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2022-2025 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

from pprint import pprint

from antlr4 import InputStream, CommonTokenStream, ParseTreeWalker
from .frontend.CypherLexer import CypherLexer
from .frontend.CypherParser import CypherParser
from .frontend.CypherListener import CypherListener

class Xlater: pass

class QueryTranslator(object):
    """
    It is sometimes necessary to make alterations to naming of graph components
    as part of the automatic graph schema creation from a Neo4j database to 
    hold data in a Rocketgraph xGT server.  One example of this is when a
    relationship type consists of some edges from one source node label and
    other edges from a different source node label.

    Given a Cypher query that formulated to run against a Neo4j database, there
    may be some changes required in order to run that same query against a
    Rocketgraph xGT server holding a graph schema that has been auto-generated.

    This class provides the service of translating a Neo4j-targeted query into
    a Rocketgraph xGT-targeted query.
    """
    def __init__(self, xgt_schemas, verbose = False):
        """
        Initializes a QueryTranslator instance.

        Parameters
        ----------
        xgt_schemas : dict
            Information about an auto-generated schema for Rocketgraph xGT
            based on a Neo4j database.  The structure of this schema should
            be the same the data returned from the
            :py:meth:`~Neo4jConnector.get_xgt_schemas` method.
        verbose : bool
            Print detailed information.  Default is False
        """
        self._verbose = verbose
        self._edge_name_mapping = self._create_edge_name_mapping(xgt_schemas)

    def _create_edge_name_mapping(self, xlate_structure) -> dict():
        mapping = dict()
        for edge, values in xlate_structure['edges'].items():
            if len(values) > 1:
                res = []
                for value in values:
                    res.append({
                        "source" : value['source'],
                        "target" : value['target'],
                        "mapped_name" : f"{value['xgt_source']}_{edge}_{value['xgt_target']}",
                        })
                mapping[edge] = res
        if self._verbose:
            print("Edge Name Mapping:")
            pprint(mapping)
        return mapping
    
    def translate(self, query:str) -> str:
        """
        Translates a Cypher query from a Neo4j-targeted query to a Rocketgraph
        xGT-targeted query.

        Parameters
        ----------
        query : str
          A Cypher query that can be run against the Neo4j database that is
          part of this connector instance.

        Returns
        -------
        str
          Translated Cypher query.
        """
        lexer = CypherLexer(InputStream(query))
        stream = CommonTokenStream(lexer)
        parser = CypherParser(stream)
        tree = parser.oC_Cypher()

        walker = ParseTreeWalker()
        xlater = Xlater(verbose = self._verbose)
        walker.walk(xlater, tree)

        self._rewrites = dict()
        for edge in xlater.edges:
            self._add_rewrites_for_edge_frame_names(edge)

        new_query = self._rewrite_query(query)
        del self._rewrites
        return new_query

    def _add_rewrites_for_edge_frame_names(self, edge) -> None:
        if self._verbose:
            print("In _add_rewrites_for_edge_frame_names")
            pprint(edge)
        left_nodes = edge[0]
        relationship = edge[1]
        right_nodes = edge[2]
        if 'rel_types' not in relationship:
            return
        for frame_name in relationship['rel_types'].keys():
            if self._verbose:
                print(f"Exploring frame name: {frame_name}")
            if frame_name in self._edge_name_mapping:
                left_node_labels = left_nodes['labels']
                if len(left_node_labels) == 0:
                    source = None
                elif len(left_node_labels) == 1:
                    source = list(left_node_labels.keys())[0]
                else:
                    raise ValueError(f"Support for multi-label left nodes not yet support: {left_nodes}")
                right_node_labels = right_nodes['labels']
                if len(right_node_labels) == 0:
                    target = None
                elif len(right_node_labels) == 1:
                    target = list(right_node_labels.keys())[0]
                else:
                    raise ValueError(f"Support for multi-label nodes not yet support: {right_nodes}")

                swap_source_target = relationship['left_arrow'] and not relationship['right_arrow']
                if swap_source_target:
                    temp = source
                    source = target
                    target = temp

                if self._verbose:
                    print(f"----> Source: {source}")
                    print(f"----> Target: {target}")

                if source is not None and target is not None:
                    mapping = self._edge_name_mapping[frame_name]
                    if self._verbose:
                        print(f"Frame to adapt: (:{source})-[:{frame_name}]->(:{target})")
                        print(f"  Frame Name {frame_name} needs adapting, {len(mapping)} mapped frames")
                    for variant in mapping:
                        if variant['source'] == source and variant['target'] == target:
                            if self._verbose:
                                print(f"Use this variant: {variant}")
                            reltype_loc = relationship['rel_types'][frame_name].location
                            if reltype_loc['offset'] in self._rewrites:
                                raise ValueError(f"Internal Error:  Multiple rewrites at the same offset ({reltype_loc['offset']}) is ambiguous")
                            self._rewrites[reltype_loc['offset']] = {
                                'from':reltype_loc, 
                                'to':variant['mapped_name']}

    def _rewrite_query(self, query:str) -> str:
        """Produce a re-written query from the self._rewrites information."""
        if self._verbose:
            print(f"Rewrites:")
            pprint(self._rewrites)
        locations = list(self._rewrites.keys())
        locations.sort(reverse = True)
        for offset in locations:
            rewrite = self._rewrites[offset]
            length = rewrite['from']['length']
            new_text = rewrite['to']
            query = f"{query[:offset]}{new_text}{query[offset+length:]}"

        return query


class QueryElement(object):
    """
    Utility object to store an element of a query.

    The original text is available as a property, and the location within
    the original query is captured.  The location includes the byte offset
    into the query, the length of the element (in bytes), and the line
    number and column number of the start and the stop positions of the
    element.
    """
    def __init__(self, text:str, location):
        self._text = text
        self._location = location

    @property
    def text(self):
        return self._text
    @property
    def location(self):
        return self._location

    def __str__(self) -> str:
        return self._text
    def __repr__(self) -> str:
        return f"<QueryElement: [{self._text}], location: {self._location}"

class Xlater(CypherListener):
    def __init__(self, verbose = False):
        self._verbose = verbose
        self._AST_node_stack = []
        self._symbol_table_node_variables = dict()
        self._edges = []
        self._node_patterns = []
        self._relationship_patterns = []

    @property
    def edges(self):
        return self._edges
    @property
    def node_patterns(self):
        return self._node_patterns

    # Capture edges in patterns
    def exitOC_PatternElementChain(self, ctx:CypherParser.OC_PatternElementChainContext):
        self._trace_context(ctx, "OC_PatternElementChain")
        left_node = self._node_patterns[-2]
        right_node = self._node_patterns[-1]
        self._edges.append([left_node, self._relationship_patterns[-1], right_node])
        if self._verbose:
            print(f"  -->Push PatternElement Edge:")
            pprint(self._edges[-1])

    def exitOC_NodePattern(self, ctx:CypherParser.OC_NodePatternContext):
        self._trace_context(ctx, "OC_NodePattern")

        if ctx.oC_NodeLabels() is None:
            labels = []
        else:
            labels = self._AST_node_stack.pop()
        variable = self._nonterminal_or_none(ctx.oC_Variable())
        if len(labels) == 0 and variable in self._symbol_table_node_variables:
            labels = self._symbol_table_node_variables[variable]
        capture = {'fulltext':ctx.getText(), 'location':self._encode_location(ctx),
                    'variable':variable, 'labels':labels}
        if variable is not None:
            if self._verbose:
                print(f"symbol: {variable}, labels: {labels}")
                if variable in self._symbol_table_node_variables:
                    print(f"Variable {variable} already in symbol table: {self._symbol_table_node_variables[variable]}")
            self._symbol_table_node_variables[variable] = labels
        self._node_patterns.append(capture)
        if self._verbose:
            print(f"  -->Push NodePattern: {capture}")

    def exitOC_NodeLabels(self, ctx:CypherParser.OC_NodeLabelsContext):
        self._trace_context(ctx, "OC_NodeLabels")
        node_labels = dict()
        for child in ctx.children:
            if isinstance(child, CypherParser.OC_NodeLabelContext):
                node_label = self._AST_node_stack.pop()
                node_labels[node_label.text] = node_label
        self._AST_node_stack.append(node_labels)
        if self._verbose:
            print(f"  -->Push NLs: {node_labels}")

    def exitOC_NodeLabel(self, ctx:CypherParser.OC_NodeLabelContext):
        self._trace_context(ctx, "OC_NodeLabel")
        node_label = self._capture_element(ctx.oC_LabelName())
        self._AST_node_stack.append(node_label)
        if self._verbose:
            print(f"  -->Push NL: {node_label}")

    def exitOC_RelationshipPattern(self, ctx:CypherParser.OC_RelationshipPatternContext):
        self._trace_context(ctx, "OC_RelationshipPattern")
        if len(self._AST_node_stack) == 0:
            rel_detail = dict()
        else:
            rel_detail = self._AST_node_stack.pop()

        left_arrow = False
        right_arrow = False
        for child in ctx.children:
            if isinstance(child, CypherParser.OC_LeftArrowHeadContext):
                left_arrow = True
            if isinstance(child, CypherParser.OC_RightArrowHeadContext):
                right_arrow = True
        capture = {'fulltext':ctx.getText(), 'location':self._encode_location(ctx),
                   'left_arrow':left_arrow, 'right_arrow':right_arrow,
                  }
        for key in rel_detail.keys():
            if key == 'location':
                capture['detail_location'] = rel_detail[key]
            else:
                capture[key] = rel_detail[key]

        self._relationship_patterns.append(capture)
        if self._verbose:
            print(f"  -> Push RelationshipPattern: {capture}")

    def exitOC_RelationshipDetail(self, ctx:CypherParser.OC_RelationshipDetailContext):
        self._trace_context(ctx, "OC_RelationshipDetail")
        if ctx.oC_RelationshipTypes() is None:
            rel_types = dict()
        else:
            rel_types = self._AST_node_stack.pop()
        capture = {'fulltext':ctx.getText(), 'location':self._encode_location(ctx),
                    'variable':self._nonterminal_or_none(ctx.oC_Variable()),
                    'rel_types':rel_types,
                    'range_literal':self._nonterminal_or_none(ctx.oC_RangeLiteral()),
                  }
        self._AST_node_stack.append(capture)
        if self._verbose:
            print(f"  -> Push RelationshipDetail: {capture}")

    def exitOC_RelationshipTypes(self, ctx:CypherParser.OC_RelationshipTypesContext):
        self._trace_context(ctx, "exitOC_RelationshipTypes")
        reltypes = dict()
        for child in ctx.children:
            if isinstance(child, CypherParser.OC_RelTypeNameContext):
                reltype = self._AST_node_stack.pop()
                reltypes[reltype.text] = reltype
        self._AST_node_stack.append(reltypes)
        if self._verbose:
            print(f"  -->Push Relationship Types: {reltypes}")

    def exitOC_RelTypeName(self, ctx:CypherParser.OC_RelTypeNameContext):
        self._trace_context(ctx, "exitOC_RelTypeName")
        reltype = self._capture_element(ctx.oC_SchemaName())
        self._AST_node_stack.append(reltype)
        if self._verbose:
            print(f"  -->Push RelTypeName: {reltype}")


    # ----------------------------------------------------------
    # Utility methods

    def _nonterminal_or_none(self, nonterminal):
        if nonterminal is None:
            return None
        return nonterminal.getText()

    def _filter_out_spaces(self, children):
        return [_ for _ in children if _.getText() != ' ']

    def _extract_specific_types_of_children(self, children, specific_type):
        return [_ for _ in children if isinstance(_, specific_type)]

    def _capture_element(self, ctx):
        if ctx is None:
            return QueryElement(None, None)
        return QueryElement(ctx.getText(), self._encode_location(ctx))

    def _encode_location(self, ctx):
        start = ctx.start
        stop = ctx.stop
        length = ctx.stop.stop - ctx.start.start + 1
        data = {'offset': start.start, 'length': length,
                'start': {'column': start.column, 'line': start.line},
                'stop': {'column': stop.column, 'line': stop.line},
               }
        return data

    def _trace_context(self, ctx, name):
        if self._verbose:
            print(f"\n====================> {name}: {ctx.getText()}")
            print(f"  getChildCount: {ctx.getChildCount()}")
            print(f"  children: {ctx.children}")



#!/usr/bin/env python
# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2022-2025 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

import unittest
from parameterized import parameterized_class
import time

import neo4j
import xgt
from xgt_connector import Neo4jConnector, Neo4jDriver

@parameterized_class([
   { "driver": "neo4j" },
   { "driver": "neo4j-bolt" },
])
class TestXgtNeo4jConnector(unittest.TestCase):
  # Print all diffs on failure.
  maxDiff=None
  @classmethod
  def setup_class(cls):
    # Create a connection to Rocketgraph xGT
    cls.xgt = xgt.Connection()
    try:
      cls.xgt.create_namespace('test')
    except:
      pass
    cls.xgt.set_default_namespace('test')
    cls.neo4j_driver, cls.conn = cls._setup_connector(cls.driver)
    return

  @classmethod
  def teardown_class(cls):
    del cls.conn
    cls.xgt.drop_namespace('test', force_drop = True)
    del cls.xgt

  @classmethod
  def _setup_connector(cls, connector_type, retries = 20):
    try:
      if connector_type == "neo4j":
        driver = neo4j.GraphDatabase.driver("neo4j://localhost", auth=('neo4j', 'foo'))
      else:
        driver = Neo4jDriver(auth=('neo4j', 'foo'), driver=connector_type)
      conn = Neo4jConnector(cls.xgt, driver)
      # Validate the db can run queries.
      with conn._neo4j_driver.bolt.session() as session:
        session.run("call db.info()")
      return (conn._neo4j_driver, conn)
    except (neo4j.exceptions.ServiceUnavailable):
      print(f"Neo4j Unavailable, retries = {retries}")
      if retries > 0:
        time.sleep(3)
        return cls._setup_connector(connector_type, retries - 1)
    if connector_type == "neo4j":
        driver = neo4j.GraphDatabase.driver("neo4j://localhost", auth=('neo4j', 'foo'))
    else:
        driver = Neo4jDriver(auth=('neo4j', 'foo'), driver=connector_type)
    conn = Neo4jConnector(cls.xgt, driver)
    return (conn._neo4j_driver, conn)

  def setup_method(self, method):
    self._erase_neo4j_database()

  def teardown_method(self, method):
    self._erase_neo4j_database()

  def test_connector_creation(self) -> None:
    # Must pass at least one parameter to constructor.
    with self.assertRaises(TypeError):
      c = Neo4jConnector()

  def test_neo4j_properties(self):
    assert isinstance(self.neo4j_driver, Neo4jDriver)
    rel = self.conn.neo4j_relationship_types
    assert isinstance(rel, list)
    assert len(rel) == 0
    labels = self.conn.neo4j_node_labels
    assert isinstance(labels, list)
    assert len(labels) == 0
    props = self.conn.neo4j_property_keys
    assert isinstance(props, list)
    assert len(props) >= 0
    props = self.conn.neo4j_node_type_properties
    assert isinstance(props, list)
    assert len(props) == 0
    props = self.conn.neo4j_rel_type_properties
    assert isinstance(props, list)
    assert len(props) == 0
    nodes = self.conn.neo4j_nodes
    assert isinstance(nodes, dict)
    assert len(nodes) == 0
    edges = self.conn.neo4j_edges
    assert isinstance(edges, dict)
    assert len(edges) == 0

  def test_node_attributes(self):
    self._populate_node()
    c = self.conn
    xgt_schema = c.get_xgt_schemas(vertices=['Node'])
    print(xgt_schema)
    vertices = xgt_schema['vertices']['Node']
    schema = dict(vertices['schema'])
    assert len(schema) == 11
    assert schema['int'] == 'int'
    assert schema['real'] == 'float'
    assert schema['str'] == 'text'
    assert schema['bool'] == 'boolean'
    assert schema['date_attr'] == 'date'
    assert schema['time_attr'] == 'time'
    assert schema['datetime_attr'] == 'datetime'
    assert schema['localtime_attr'] == 'time'
    assert schema['localdatetime_attr'] == 'datetime'
    assert schema['duration_attr'] == 'int'
    print(schema)
    print(c.neo4j_node_type_properties)

  def test_neo4j_relationship_types(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE (:Node2{})-[:Relationship1{}]->(:Node1{}), (:Node1{})-[:Relationship2{int: 1}]->(:Node2{})').finalize()
    self.assertCountEqual(c.neo4j_relationship_types, ['Relationship1', 'Relationship2'])

  def test_neo4j_node_labels(self):
    c = self.conn
    self.neo4j_driver.query('CREATE (:Node1{}), (:Node2{int : 1})').finalize()
    self.assertCountEqual(c.neo4j_node_labels, ['Node1', 'Node2'])

  def test_neo4j_property_keys(self):
    c = self.conn
    self._populate_node_working_types_bolt()
    self.assertCountEqual(
        c.neo4j_property_keys,
        ['bool', 'bool_array', 'date_array', 'date_attr', 'datetime_array', 'datetime_attr', 'double_array', 'duration_array', 'duration_attr',
         'geo_attr2', 'geo_attr3', 'int', 'localdatetime_array', 'localdatetime_attr', 'localtime_array', 'localtime_attr', 'long_array',
         'point_array', 'point_attr2', 'point_attr3', 'real', 'str', 'string_array', 'time_array', 'time_attr', 'x', 'y'])

  def test_neo4j_rel_type_properties(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE (:Node2{})-[:Relationship1{}]->(:Node1{}), (:Node1{})-[:Relationship2{int: 1, str: "hello"}]->(:Node2{}),'
        '(:Node1{})-[:Relationship2{str: "goodbye"}]->(:Node2{})').finalize()
    self.assertCountEqual(
        c.neo4j_rel_type_properties,
        [{'relType': ':`Relationship1`', 'propertyName': None, 'propertyTypes': None, 'mandatory': False},
         {'relType': ':`Relationship2`', 'propertyName': 'int', 'propertyTypes': ['Long'], 'mandatory': False},
         {'relType': ':`Relationship2`', 'propertyName': 'str', 'propertyTypes': ['String'], 'mandatory': True}])

  def test_neo4j_node_type_properties(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE (:Node1{}), (:Node2{int : 1, str : "hello"}), (:Node2{str : "goodbye"})').finalize()
    self.assertCountEqual(
        c.neo4j_node_type_properties,
        [{'nodeType': ':`Node1`', 'nodeLabels': ['Node1'], 'propertyName': None, 'propertyTypes': None, 'mandatory': False},
         {'nodeType': ':`Node2`', 'nodeLabels': ['Node2'], 'propertyName': 'int', 'propertyTypes': ['Long'], 'mandatory': False},
         {'nodeType': ':`Node2`', 'nodeLabels': ['Node2'], 'propertyName': 'str', 'propertyTypes': ['String'], 'mandatory': True}])

  def test_neo4j_edges(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE (:Node2{})-[:Relationship1{}]->(:Node1{}), (:Node1{})-[:Relationship2{int: 1}]->(:Node2{})').finalize()
    assert len(c.neo4j_edges) == 2
    assert c.neo4j_edges['Relationship1']['endpoints'] == {('Node2', 'Node1')}
    assert c.neo4j_edges['Relationship1']['sources'] == {'Node2'}
    assert c.neo4j_edges['Relationship1']['targets'] == {'Node1'}
    assert c.neo4j_edges['Relationship1']['schema'] == {}
    assert c.neo4j_edges['Relationship2']['endpoints'] == {('Node1', 'Node2')}
    assert c.neo4j_edges['Relationship2']['sources'] == {'Node1'}
    assert c.neo4j_edges['Relationship2']['targets'] == {'Node2'}
    assert c.neo4j_edges['Relationship2']['schema'] == {'int' : 'Long'}

  def test_neo4j_edges_multi(self):
    c = self.conn
    self.neo4j_driver.query(
          'CREATE (:Node2{})-[:Relationship1{int: 2}]->(:Node1{}), (:Node1{})-[:Relationship1{int: 1}]->(:Node2{})').finalize()
    assert len(c.neo4j_edges) == 1
    assert c.neo4j_edges['Relationship1']['endpoints'] == {('Node1', 'Node2'), ('Node2', 'Node1')}
    assert c.neo4j_edges['Relationship1']['sources'] == {'Node1', 'Node2'}
    assert c.neo4j_edges['Relationship1']['targets'] == {'Node1', 'Node2'}
    assert c.neo4j_edges['Relationship1']['schema'] == {'int' : 'Long'}

  def test_neo4j_nodes(self):
    c = self.conn
    self.neo4j_driver.query('CREATE (:Node1{}), (:Node2{int : 1})').finalize()
    assert len(c.neo4j_nodes) == 2
    assert c.neo4j_nodes['Node1'] == {}
    assert c.neo4j_nodes['Node2'] == {'int' : 'Long'}

  def test_graph_update_after_connector_created(self):
    c = self.conn
    self._populate_node()
    xgt_schema = c.get_xgt_schemas(vertices=['Node'])
    vertices = xgt_schema['vertices']['Node']
    schema = dict(vertices['schema'])
    assert len(schema) == 11
    assert schema['int'] == 'int'
    assert schema['real'] == 'float'
    assert schema['str'] == 'text'
    assert schema['bool'] == 'boolean'
    assert schema['date_attr'] == 'date'
    assert schema['time_attr'] == 'time'
    assert schema['datetime_attr'] == 'datetime'
    assert schema['localtime_attr'] == 'time'
    assert schema['localdatetime_attr'] == 'datetime'
    assert schema['duration_attr'] == 'int'

  def test_graph_delete_after_connector_created(self):
    self._populate_node()
    c = self.conn
    self.neo4j_driver.query("MATCH (n) DETACH DELETE n").finalize()
    with self.assertRaises(ValueError):
      xgt_schema = c.get_xgt_schemas(vertices=['Node'])

  def disable_test_transfer_node(self):
    self._populate_node()
    c = self.conn
    xgt_schema = c.get_xgt_schemas(vertices=['Node'])
    c.create_xgt_schemas(xgt_schema)
    for vertex, schema in xgt_schema['vertices'].items():
      table_schema = schema['schema']
      attributes = {_:t for _, t in table_schema}
      print(f"\nAttributes: {attributes}")
    c.copy_data_to_xgt(xgt_schema)
    node_frame = self.xgt.get_frame('Node')
    assert node_frame.num_rows == 1

  def test_transfer_node_working_types_bolt(self):
    self._populate_node_working_types_bolt()
    c = self.conn
    xgt_schema = c.get_xgt_schemas(vertices=['Node'])
    c.create_xgt_schemas(xgt_schema)
    for vertex, schema in xgt_schema['vertices'].items():
      table_schema = schema['schema']
      attributes = {_:t for _, *t  in table_schema}
      print(f"\nAttributes: {attributes}")
    c.copy_data_to_xgt(xgt_schema)
    node_frame = self.xgt.get_frame('Node')
    assert node_frame.num_rows == 3
    print(node_frame.get_data())

  def test_transfer_nodes_to_neo4j(self):
    self._populate_node_working_types_bolt()
    c = self.conn
    c.transfer_to_xgt(vertices=['Node'])
    node_frame = self.xgt.get_frame('Node')
    assert node_frame.num_rows == 3
    expected = [row[1:] for row in node_frame.get_data()]
    self.neo4j_driver.query("MATCH (n) DETACH DELETE n").finalize()
    c.transfer_to_neo4j(vertices=['Node'])
    self.xgt.drop_frame("Node")
    c.transfer_to_xgt(vertices=['Node'])
    node_frame = self.xgt.get_frame('Node')
    assert node_frame.num_rows == 3
    result = [row[1:] for row in node_frame.get_data()]
    # The column ordering can change when uploading to neo4j
    # So we just need to verify the None rows are correct
    # and the same elements are in the row with data, but with a different order.
    expected_pos = 0
    result_pos = 0
    for i in range(0,3):
        if result[i][0] != None:
            result_pos = i
        else:
            assert result[i].count(None) == len(result[i])
        if expected[i][0] != None:
            expected_pos = i
        else:
            assert expected[i].count(None) == len(expected[i])
    self.assertCountEqual(expected[expected_pos], result[result_pos])

  def test_transfer_edges_to_neo4j(self):
    self._populate_relationship_working_types_bolt()
    c = self.conn
    c.transfer_to_xgt(edges=['Relationship'])
    node_frame = self.xgt.get_frame('Node')
    edge_frame = self.xgt.get_frame('Relationship')
    assert edge_frame.num_rows == 3
    node_expected = [row[1:] for row in node_frame.get_data()]
    edge_expected = [row[2:] for row in edge_frame.get_data()]
    self.neo4j_driver.query("MATCH (n) DETACH DELETE n").finalize()
    c.transfer_to_neo4j(edges=['Relationship'])
    self.xgt.drop_frame("Relationship")
    self.xgt.drop_frame("Node")
    c.transfer_to_xgt(edges=['Relationship'])
    node_frame = self.xgt.get_frame('Node')
    edge_frame = self.xgt.get_frame('Relationship')
    assert edge_frame.num_rows == 3
    node_result = [row[1:] for row in node_frame.get_data()]
    edge_result = [row[2:] for row in edge_frame.get_data()]
    self.assertCountEqual(node_expected, node_result)
    # The column ordering can change when uploading to neo4j
    # So we just need to verify the None rows are correct
    # and the same elements are in the row with data, but with a different order.
    expected_pos = 0
    result_pos = 0
    for i in range(0,3):
        if edge_result[i][0] != None:
            result_pos = i
        else:
            assert edge_result[i].count(None) == len(edge_result[i])
        if edge_expected[i][0] != None:
            expected_pos = i
        else:
            assert edge_expected[i].count(None) == len(edge_expected[i])
    self.assertCountEqual(edge_expected[expected_pos], edge_result[result_pos])

  def test_transfer_relationship_working_types_bolt(self):
    self._populate_relationship_working_types_bolt()
    c = self.conn
    xgt_schema = c.get_xgt_schemas(vertices=['Node'], edges=['Relationship'])
    c.create_xgt_schemas(xgt_schema)
    c.copy_data_to_xgt(xgt_schema)
    edge_frame = self.xgt.get_frame('Relationship')
    assert edge_frame.num_rows == 3
    print(edge_frame.get_data())
    self.xgt.drop_frame("Relationship")

  def test_transfer_relationship_working_types_bolt_map(self):
    self._populate_relationship_working_types_bolt()
    c = self.conn
    xgt_schema = c.get_xgt_schemas(vertices=[('Node', 'n')], edges=[('Relationship', 'r')])
    c.create_xgt_schemas(xgt_schema)
    c.copy_data_to_xgt(xgt_schema)
    node_frame = self.xgt.get_frame('n')
    edge_frame = self.xgt.get_frame('r')
    assert node_frame.num_rows == 6
    assert edge_frame.num_rows == 3
    print(edge_frame.get_data())
    self.xgt.drop_frame('r')

  def test_transfer_relationship_without_vertex_bolt(self):
    self._populate_relationship_working_types_bolt()
    c = self.conn
    xgt_schema = c.get_xgt_schemas(edges=['Relationship'])
    c.create_xgt_schemas(xgt_schema)
    c.copy_data_to_xgt(xgt_schema)
    node_frame = self.xgt.get_frame('Node')
    edge_frame = self.xgt.get_frame('Relationship')
    assert node_frame.num_rows == 6
    assert edge_frame.num_rows == 3
    assert node_frame.get_data(length=1)[0][1] == 1
    self.xgt.drop_frame("Relationship")

  def test_transfer_relationship_without_vertex_bolt_two_phase(self):
    self._populate_relationship_working_types_bolt()
    c = self.conn
    c.transfer_to_xgt(vertices=['Node'])
    xgt_schema = c.get_xgt_schemas(edges=['Relationship'], import_edge_nodes=False)
    c.create_xgt_schemas(xgt_schema, append=True)
    c.copy_data_to_xgt(xgt_schema)
    node_frame = self.xgt.get_frame('Node')
    edge_frame = self.xgt.get_frame('Relationship')
    assert node_frame.num_rows == 6
    assert edge_frame.num_rows == 3
    assert node_frame.get_data(length=1)[0][1] == 1
    self.xgt.drop_frame("Relationship")

  def test_transfer_everything_bolt(self):
    self._populate_relationship_working_types_bolt()
    c = self.conn
    xgt_schema = c.get_xgt_schemas()
    c.create_xgt_schemas(xgt_schema)
    c.copy_data_to_xgt(xgt_schema)
    node_frame = self.xgt.get_frame('Node')
    edge_frame = self.xgt.get_frame('Relationship')
    assert node_frame.num_rows == 6
    assert edge_frame.num_rows == 3
    self.xgt.drop_frame("Relationship")

  def test_append(self):
    c = self.conn
    self.neo4j_driver.query(
      'CREATE (:Node{int: 343, str: "string"})').finalize()
    xgt_schema = c.get_xgt_schemas(vertices=['Node'])
    c.create_xgt_schemas(xgt_schema)

    c.copy_data_to_xgt(xgt_schema)
    c.create_xgt_schemas(xgt_schema, append=True)

    self.neo4j_driver.query("MATCH (n) DETACH DELETE n").finalize()
    self.neo4j_driver.query(
        'CREATE (:Node{int: 344, str: "string"})').finalize()
    c.copy_data_to_xgt(xgt_schema)
    node_frame = self.xgt.get_frame('Node')
    assert node_frame.num_rows == 2

    c.create_xgt_schemas(xgt_schema, append=False)
    node_frame = self.xgt.get_frame('Node')
    assert node_frame.num_rows == 0

  def test_dropping(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE (:Node{})-[:Relationship]->(:Node{})').finalize()
    xgt_schema1 = c.get_xgt_schemas(vertices=['Node'], edges=['Relationship'])
    xgt_schema2 = c.get_xgt_schemas(vertices=['Node'])

    c.create_xgt_schemas(xgt_schema1)
    self.xgt.get_frame('Node')
    self.xgt.get_frame('Relationship')

    c.create_xgt_schemas(xgt_schema1)
    self.xgt.get_frame('Node')
    self.xgt.get_frame('Relationship')

    with self.assertRaises(xgt.XgtFrameDependencyError):
        c.create_xgt_schemas(xgt_schema2)
    c.create_xgt_schemas(xgt_schema2, force=True)
    self.xgt.get_frame('Node')
    with self.assertRaises(xgt.XgtNameError):
        self.xgt.get_frame('Relationship')

  def test_multiple_node_labels_to(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE (:Node1{})-[:Relationship{}]->(:Node1{}), (:Node1{})-[:Relationship{}]->(:Node2{})').finalize()
    schema = c.get_xgt_schemas(vertices=['Node1', 'Node2'], edges=['Relationship'])
    c.create_xgt_schemas(schema)
    c.copy_data_to_xgt(schema)

    node_frame = self.xgt.get_frame('Node1_Relationship_Node1')
    assert node_frame.num_rows == 1
    node_frame = self.xgt.get_frame('Node1_Relationship_Node2')
    assert node_frame.num_rows == 1
    self.xgt.drop_frame("Node1_Relationship_Node1")
    self.xgt.drop_frame("Node1_Relationship_Node2")

  def test_multiple_node_labels_from(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE (:Node1{})-[:Relationship{}]->(:Node1{}), (:Node2{})-[:Relationship{}]->(:Node1{})').finalize()
    schema = c.get_xgt_schemas(vertices=['Node1', 'Node2'], edges=['Relationship'])
    c.create_xgt_schemas(schema)
    c.copy_data_to_xgt(schema)

    node_frame = self.xgt.get_frame('Node1_Relationship_Node1')
    assert node_frame.num_rows == 1
    node_frame = self.xgt.get_frame('Node2_Relationship_Node1')
    assert node_frame.num_rows == 1
    self.xgt.drop_frame("Node1_Relationship_Node1")
    self.xgt.drop_frame("Node2_Relationship_Node1")

  def test_multiple_property_types_vertex_negative(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE (:Node{x: 1})-[:Relationship{}]->(:Node{x: "hello"})').finalize()
    with self.assertRaises(ValueError):
        c.get_xgt_schemas(vertices=['Node'], edges=['Relationship'])

  def test_multiple_property_types_edge_negative(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE (:Node{})-[:Relationship{x: 1}]->(:Node{}), (:Node{})-[:Relationship{x: "hello"}]->(:Node{})').finalize()
    with self.assertRaises(ValueError):
      c.get_xgt_schemas(vertices=['Node'], edges=['Relationship'])

  def test_different_properties_combine_into_single_schema(self):
    c = self.conn
    self.neo4j_driver.query('CREATE (:Node{x: 1}), (:Node{y: "hello"})').finalize()
    schema = c.get_xgt_schemas(vertices=['Node'])
    node_schema = schema['vertices']['Node']['schema']
    assert len(node_schema) == 3

  def test_transfer_no_data(self):
    c = self.conn
    # Transfer all graphs:
    c.transfer_to_xgt()
    # Transfer all graphs in namespace:
    c.transfer_to_neo4j()

  def test_transfer_missing_graphs(self):
    c = self.conn
    with self.assertRaises(ValueError):
        c.transfer_to_xgt(vertices=['cars'])
    with self.assertRaises(ValueError):
        c.transfer_to_xgt(edges=['cars'])
    with self.assertRaises(ValueError):
        c.get_xgt_schemas(vertices=['cars'])
    with self.assertRaises(ValueError):
        c.get_xgt_schemas(edges=['cars'])
    with self.assertRaises(xgt.XgtNameError):
        c.transfer_to_neo4j(vertices=['cars'])
    with self.assertRaises(xgt.XgtNameError):
        c.transfer_to_neo4j(edges=['cars'])

  def test_empty_labels(self):
    c = self.conn
    self.neo4j_driver.query('CREATE ()').finalize()
    c.transfer_to_xgt(vertices=[''])
    node_frame = self.xgt.get_frame('unlabeled')
    assert node_frame.num_rows == 1

  def test_map_empty_labels(self):
    c = self.conn
    self.neo4j_driver.query('CREATE ()').finalize()
    c.transfer_to_xgt(vertices=[('', 'custom_name')])
    node_frame = self.xgt.get_frame('custom_name')
    assert node_frame.num_rows == 1
    with self.assertRaises(xgt.XgtNameError):
        node_frame = self.xgt.get_frame('unlabeled')

  def test_empty_labels_schema(self):
    c = self.conn
    self.neo4j_driver.query('CREATE ({int:2})').finalize()
    c.transfer_to_xgt(vertices=[''])
    node_frame = self.xgt.get_frame('unlabeled')
    res = self.xgt.run_job('match (v) return v.int')
    assert node_frame.num_rows == 1
    assert res.get_data()[0][0] == 2
    self.neo4j_driver.query("MATCH (n) DETACH DELETE n").finalize()
    self.neo4j_driver.query('CREATE ({str:"bbb"})').finalize()
    c.transfer_to_xgt(vertices=[''])
    node_frame = self.xgt.get_frame('unlabeled')
    res = self.xgt.run_job('match (v) return v.str')
    assert node_frame.num_rows == 1
    assert res.get_data()[0][0] == "bbb"

  def test_edge_empty_labels(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE ()-[:e1]->(), (:Node)-[:e2]->(), ()-[:e3]->(:Node)').finalize()
    c.transfer_to_xgt(edges=['e1', 'e2', 'e3'])
    node_frame = self.xgt.get_frame('e1')
    assert node_frame.num_rows == 1
    node_frame = self.xgt.get_frame('e2')
    assert node_frame.num_rows == 1
    node_frame = self.xgt.get_frame('e3')
    assert node_frame.num_rows == 1
    node_frame = self.xgt.get_frame('unlabeled')
    assert node_frame.num_rows == 4
    node_frame = self.xgt.get_frame('Node')
    assert node_frame.num_rows == 2

  def test_edge_empty_labels_schema(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE ()-[:e1]->({int:1}), (:Node)-[:e2]->({str:"aaa"}), ({bool:True})-[:e3]->(:Node)').finalize()
    c.transfer_to_xgt(edges=['e1', 'e2', 'e3'])
    node_frame = self.xgt.get_frame('e1')
    res = self.xgt.run_job('match ()-[:e1]->(v) return v.int')
    assert node_frame.num_rows == 1
    assert res.get_data()[0][0] == 1
    node_frame = self.xgt.get_frame('e2')
    res = self.xgt.run_job('match ()-[:e2]->(v) return v.str')
    assert node_frame.num_rows == 1
    assert res.get_data()[0][0] == "aaa"
    node_frame = self.xgt.get_frame('e3')
    res = self.xgt.run_job('match (v)-[:e3]->() return v.bool')
    assert node_frame.num_rows == 1
    assert res.get_data()[0][0] == True
    node_frame = self.xgt.get_frame('unlabeled')
    assert node_frame.num_rows == 4
    node_frame = self.xgt.get_frame('Node')
    assert node_frame.num_rows == 2

  def test_map_edge_empty_labels(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE ()-[:e1]->(), (:Node)-[:e2]->(), ()-[:e3]->(:Node)').finalize()
    c.transfer_to_xgt(vertices=[('', 'custom_empty'), ('Node', 'custom_Node')], edges=[('e1', 'edge1'), ('e2', 'edge2'), 'e3'])
    node_frame = self.xgt.get_frame('edge1')
    assert node_frame.num_rows == 1
    node_frame = self.xgt.get_frame('edge2')
    assert node_frame.num_rows == 1
    node_frame = self.xgt.get_frame('e3')
    assert node_frame.num_rows == 1
    node_frame = self.xgt.get_frame('custom_empty')
    assert node_frame.num_rows == 4
    node_frame = self.xgt.get_frame('custom_Node')
    assert node_frame.num_rows == 2
    with self.assertRaises(xgt.XgtNameError):
        node_frame = self.xgt.get_frame('unlabeled')
    with self.assertRaises(xgt.XgtNameError):
        node_frame = self.xgt.get_frame('Node')

  def test_edge_empty_labels_schema_no_implicit(self):
    c = self.conn
    self.neo4j_driver.query(
        'CREATE ()-[:e1]->({int:1}), (:Node)-[:e2]->({str:"aaa"}), ({bool:True})-[:e3]->(:Node)').finalize()
    c.transfer_to_xgt(vertices=['', 'Node'])
    c.transfer_to_xgt(edges=['e1', 'e2', 'e3'], append=True, import_edge_nodes=False)
    node_frame = self.xgt.get_frame('e1')
    res = self.xgt.run_job('match ()-[:e1]->(v) return v.int')
    assert node_frame.num_rows == 1
    assert res.get_data()[0][0] == 1
    node_frame = self.xgt.get_frame('e2')
    res = self.xgt.run_job('match ()-[:e2]->(v) return v.str')
    assert node_frame.num_rows == 1
    assert res.get_data()[0][0] == "aaa"
    node_frame = self.xgt.get_frame('e3')
    res = self.xgt.run_job('match (v)-[:e3]->() return v.bool')
    assert node_frame.num_rows == 1
    assert res.get_data()[0][0] == True
    node_frame = self.xgt.get_frame('unlabeled')
    assert node_frame.num_rows == 4
    node_frame = self.xgt.get_frame('Node')
    assert node_frame.num_rows == 2

  def test_query_translator_loop_data(self):
    self.neo4j_driver.query("create (a:Node{int:1})-[b:EDGE{int:1}]->(a)").finalize()
    query = "match(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node) return a"
    assert query == self.conn.translate_query(query)
    query = "match(v0)-[]->(v1) return v0"
    assert query == self.conn.translate_query(query)
    query = "match(v0:Node)-[]->(v1) return v0"
    assert query == self.conn.translate_query(query)
    query = "match(v0:Node)-[]->(v1:Node) return v0"
    assert query == self.conn.translate_query(query)
    query = "match(v0:Node)-[e:EDGE]->(v1:Node) return v0"
    assert query == self.conn.translate_query(query)
    query = "match()-[e]->() return v0"
    assert query == self.conn.translate_query(query)
    query = "match()-[e:EDGE]->() return v0"
    assert query == self.conn.translate_query(query)
    query = "match(a)-->()-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a)-->()-->()-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node)-->()-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node)-->()-->()-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node)-->(e:Edge)-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node)-->(e1:Edge)-->(e2:Edge)-->(a) return a"
    assert query == self.conn.translate_query(query)

  def test_query_translator_loop_no_data(self):
    query = "match(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node) return a"
    assert query == self.conn.translate_query(query)
    query = "match(v0)-[]->(v1) return v0"
    assert query == self.conn.translate_query(query)
    query = "match(v0:Node)-[]->(v1) return v0"
    assert query == self.conn.translate_query(query)
    query = "match(v0:Node)-[]->(v1:Node) return v0"
    assert query == self.conn.translate_query(query)
    query = "match(v0:Node)-[e:EDGE]->(v1:Node) return v0"
    assert query == self.conn.translate_query(query)
    query = "match()-[e]->() return v0"
    assert query == self.conn.translate_query(query)
    query = "match()-[e:EDGE]->() return v0"
    assert query == self.conn.translate_query(query)
    query = "match(a)-->()-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a)-->()-->()-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node)-->()-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node)-->()-->()-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node)-->(e:Edge)-->(a) return a"
    assert query == self.conn.translate_query(query)
    query = "match(a:Node)-->(e1:Edge)-->(e2:Edge)-->(a) return a"
    assert query == self.conn.translate_query(query)

  def test_query_translator_multiple_node_labels_to(self):
    c = self.conn
    self.neo4j_driver.query(
      'CREATE (:Node1{})-[:REL{}]->(:Node1{}), (:Node1{})-[:REL{}]->(:Node2{})').finalize()

    query = "MATCH (:Node1)-[:REL]->(b:Node1) RETURN count(*)"
    assert "Node1_REL_Node1" in c.translate_query(query)
    query = "MATCH (:Node1)-[:REL]->(b:Node2) RETURN count(*)"
    assert "Node1_REL_Node2" in c.translate_query(query)
    query = "MATCH (:Node2)-[:REL]->(b:Node1) RETURN count(*)"
    assert "Node2_REL_Node1" not in c.translate_query(query)
    query = "MATCH (:Node1)-[:REL]->(b:Node3) RETURN count(*)"
    assert "Node1_REL_Node3" not in c.translate_query(query)

  def test_query_translator_multiple_node_labels_from(self):
    c = self.conn
    self.neo4j_driver.query(
      'CREATE (:Node1{})-[:REL{}]->(:Node1{}), (:Node2{})-[:REL{}]->(:Node1{})').finalize()

    query = "MATCH (:Node1)-[:REL]->(b:Node1) RETURN count(*)"
    assert "Node1_REL_Node1" in c.translate_query(query)
    query = "MATCH (:Node2)-[:REL]->(b:Node1) RETURN count(*)"
    assert "Node2_REL_Node1" in c.translate_query(query)
    query = "MATCH (:Node1)-[:REL]->(b:Node2) RETURN count(*)"
    assert "Node1_REL_Node2" not in c.translate_query(query)
    query = "MATCH (:Node3)-[:REL]->(b:Node1) RETURN count(*)"
    assert "Node3_REL_Node1" not in c.translate_query(query)

  def test_query_translator_multiple_node_labels_to_and_from(self):
    c = self.conn
    self.neo4j_driver.query(
      'CREATE (:Node1{})-[:REL{}]->(:Node1{}), (:Node2{})-[:REL{}]->(:Node1{}),'
            '(:Node1{})-[:REL{}]->(:Node2{}), (:Node2{})-[:REL{}]->(:Node2{})').finalize()

    query = "MATCH (:Node1)-[:REL]->(b:Node1) RETURN count(*)"
    assert "Node1_REL_Node1" in c.translate_query(query)
    query = "MATCH (:Node2)-[:REL]->(b:Node1) RETURN count(*)"
    assert "Node2_REL_Node1" in c.translate_query(query)
    query = "MATCH (:Node1)-[:REL]->(b:Node2) RETURN count(*)"
    assert "Node1_REL_Node2" in c.translate_query(query)
    query = "MATCH (:Node3)-[:REL]->(b:Node1) RETURN count(*)"
    assert "Node3_REL_Node1" not in c.translate_query(query)
    query = "MATCH (:Node1)-[:REL]->(b:Node3) RETURN count(*)"
    assert "Node1_REL_Node3" not in c.translate_query(query)

  def _populate_node(self):
    self.neo4j_driver.query(
      # Integer, Float, String, Boolean, Point, Date, Time, LocalTime,
      # DateTime, LocalDateTime, and Duration.
      # FIXME: Point listed in comment above, but not in the list
      'CREATE (:Node{int: 343, real: 3.14, str: "string", bool: true, ' +
      'date_attr: date("+2015-W13-4"), time_attr: time("125035.556+0100"), ' +
      'datetime_attr: datetime("2015-06-24T12:50:35.556+0100"), ' +
      'localtime_attr: localtime("12:50:35.556"), ' +
      'localdatetime_attr: localdatetime("2015185T19:32:24"), ' +
      'duration_attr: duration("P14DT16H12M")})').finalize()

  # Point not working for bolt.
  def _populate_node_working_types_bolt(self):
    self.neo4j_driver.query(
      # Integer, Float, String, Boolean, Date, Time, LocalTime,
      # DateTime, LocalDateTime, various Points, and Lists.
      'CREATE (:Node{}), (:Node{int: 343, real: 3.14, str: "string", bool: true, ' +
      'date_attr: date("+2015-W13-4"), time_attr: time("125035.556+0100"), ' +
      'datetime_attr: datetime("2015-06-24T12:50:35.556+0100"), ' +
      'localtime_attr: localtime("12:50:35.556"), ' +
      'localdatetime_attr: localdatetime("2015185T19:32:24"), ' +
      'duration_attr: duration({days: 14, hours:16, minutes: 12}), ' +
      'point_attr2: point({x:0.5, y:1.2}), ' +
      'point_attr3: point({x:0.23, y:1.5, z:1.2}), ' +
      'geo_attr2: point({latitude:7.23, longitude:3.5}), ' +
      'geo_attr3: point({latitude:0.23, longitude:1.5, height:10.2}), ' +
      'bool_array: [true, false, false], ' +
      'long_array: [7, 1, 5], ' +
      'string_array: ["ad", "bc", "de"], ' +
      'double_array: [0.7, 1.9, 5.2], ' +
      'date_array: [date("+2015-W13-4"), date("+2000-W1-2"), date("+2011-W4-6")], ' +
      'time_array: [time("125035.556+0100"), time("110535.7+0500"), time("100000.0+0700")], ' +
      'datetime_array: [datetime("2015-06-24T12:50:35.556+0100"), datetime("2000-01-1T1:1:1+0400"), datetime("1904-02-1T10:40:12.34+0300")], ' +
      'localtime_array: [localtime("125035.556"), localtime("110535.7"), localtime("100000.0")], ' +
      'localdatetime_array: [localdatetime("2015-06-24T12:50:35.556"), localdatetime("2000-01-1T1:1:1"), localdatetime("1904-02-1T10:40:12.34")], ' +
      'duration_array: [duration({days: 14, hours:16, minutes: 12}), duration({days: 5, hours:5, minutes: 5}), duration({days: 0, hours:6, minutes: 1})], ' +
      'point_array: [point({x:0.5, y:1.2}), point({x:0.5, y:1.2}), point({x:0.5, y:1.2})]}), ' +
      '(:Node{})').finalize()

  # Point not working for bolt.
  def _populate_relationship_working_types_bolt(self):
    self.neo4j_driver.query(
      # Integer, Float, String, Boolean, Date, Time, LocalTime,
      # DateTime, LocalDateTime, various Points, and Lists.
      'CREATE (:Node{int: 1})-[:Relationship{}]->(:Node{int: 1}), (:Node{int: 1})-' +
      '[:Relationship{int: 343, real: 3.14, str: "string", bool: true, ' +
      'date_attr: date("+2015-W13-4"), time_attr: time("125035.556+0100"), ' +
      'datetime_attr: datetime("2015-06-24T12:50:35.556+0100"), ' +
      'localtime_attr: localtime("12:50:35.556"), ' +
      'localdatetime_attr: localdatetime("2015185T19:32:24"), ' +
      'duration_attr: duration({days: 14, hours:16, minutes: 12}),' +
      'point_attr2: point({x:0.5, y:1.2}), ' +
      'point_attr3: point({x:0.23, y:1.5, z:1.2}), ' +
      'geo_attr2: point({latitude:7.23, longitude:3.5}), ' +
      'geo_attr3: point({latitude:0.23, longitude:1.5, height:10.2}),' +
      'bool_array: [true, false, false], ' +
      'long_array: [7, 1, 5], ' +
      'string_array: ["ad", "bc", "de"], ' +
      'double_array: [0.7, 1.9, 5.2], ' +
      'date_array: [date("+2015-W13-4"), date("+2000-W1-2"), date("+2011-W4-6")], ' +
      'time_array: [time("125035.556+0100"), time("110535.7+0500"), time("100000.0+0700")], ' +
      'datetime_array: [datetime("2015-06-24T12:50:35.556+0100"), datetime("2000-01-1T1:1:1+0400"), datetime("1904-02-1T10:40:12.34+0300")], ' +
      'localtime_array: [localtime("125035.556"), localtime("110535.7"), localtime("100000.0")], ' +
      'localdatetime_array: [localdatetime("2015-06-24T12:50:35.556"), localdatetime("2000-01-1T1:1:1"), localdatetime("1904-02-1T10:40:12.34")], ' +
      'duration_array: [duration({days: 14, hours:16, minutes: 12}), duration({days: 5, hours:5, minutes: 5}), duration({days: 0, hours:6, minutes: 1})], ' +
      'point_array: [point({x:0.5, y:1.2}), point({x:0.5, y:1.2}), point({x:0.5, y:1.2})]}]' +
      '->(:Node{int: 1}), (:Node{int: 1})-[:Relationship{}]->(:Node{int: 1})').finalize()

  def _erase_neo4j_database(self):
    self.neo4j_driver.query("MATCH (n) DETACH DELETE n").finalize()
    self.xgt.drop_namespace('test', force_drop = True)



#!/usr/bin/env python
# -*- coding: utf-8 -*- --------------------------------------------------===#
#
#  Copyright 2022-2025 Trovares Inc.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#===----------------------------------------------------------------------===#

import arrow_odbc
import struct
import sys
import xgt
import pyarrow as pa
import pyarrow.flight as pf

from collections.abc import Iterable, Mapping, Sequence

Iter, Map, Seq, List, Dict = Iterable, Mapping, Sequence, list, dict

from arrow_odbc import read_arrow_batches_from_odbc
from arrow_odbc import insert_into_table
from typing import Optional, Union, TYPE_CHECKING
from xgt import SchemaMessages_pb2 as sch_proto
from .common import ProgressDisplay

# Convert the pyarrow type to an xgt type.
def _pyarrow_type_to_xgt_type(pyarrow_type):
    if pa.types.is_boolean(pyarrow_type):
        return xgt.BOOLEAN
    elif pa.types.is_timestamp(pyarrow_type) or pa.types.is_date64(pyarrow_type):
        return xgt.DATETIME
    elif pa.types.is_date(pyarrow_type):
        return xgt.DATE
    elif pa.types.is_time(pyarrow_type):
        return xgt.TIME
    elif pa.types.is_integer(pyarrow_type):
        return xgt.INT
    elif pa.types.is_float32(pyarrow_type) or \
         pa.types.is_float64(pyarrow_type) or \
         pa.types.is_decimal(pyarrow_type):
        return xgt.FLOAT
    elif pa.types.is_string(pyarrow_type):
        return xgt.TEXT
    else:
        raise xgt.XgtTypeError("Cannot convert pyarrow type " + str(pyarrow_type) + " to xGT type.")

def _infer_xgt_schema_from_pyarrow_schema(pyarrow_schema, conversions):
    schema = []
    for field in pyarrow_schema:
        if field.type in conversions:
            field = pa.field(field.name, conversions[field.type], field.nullable, field.metadata)
        schema.append(field)

    schema = pa.schema(schema)

    return [[c.name, _pyarrow_type_to_xgt_type(c.type)] for c in schema]

class SQLODBCDriver(object):
    def __init__(self, connection_string : str):
        """
        Initializes the driver class.

        Parameters
        ----------
        connection_string : str
            Standard ODBC connection string used for connecting to the ODBC applications.
            Example:
            'Driver={MariaDB};Server=127.0.0.1;Port=3306;Database=test;Uid=test;Pwd=foo;'
        """
        self._connection_string = connection_string
        self._schema_query = "SELECT * FROM {0} LIMIT 1;"
        self._data_query = "SELECT * FROM {0};"
        self._estimate_query="SELECT TABLE_ROWS FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = '{0}';"

    def _get_data_query(self, table, arrow_schema):
        return  self._data_query.format(table)

    def _conversions(self):
       return { }

    def _get_record_batch_schema(self, table, max_text_size, max_binary_size):
        reader = read_arrow_batches_from_odbc(
            query=self._schema_query.format(table),
            connection_string=self._connection_string,
            batch_size=1,
            max_text_size=max_text_size,
            max_binary_size=max_binary_size,
        )
        return reader.schema

class MongoODBCDriver(object):
    def __init__(self, connection_string : str , include_id : bool = False):
        """
        Initializes the driver class.

        Parameters
        ----------
        connection_string : str
            Standard ODBC connection string used for connecting to MongoDB.
            Example:
            'DSB=MongoDB;Database=test;Uid=test;Pwd=foo;'
        include_id : boolean
            Include the MongoDB id field when transferring from MongoDB.
            If the id field is included, writing data back to the database will update the columns
            instead of inserting new rows.
            By default false.
        """
        self._connection_string = connection_string
        self._schema_query = "SELECT * FROM {0} LIMIT 1;"
        self._data_query = "SELECT {0} FROM {1};"
        self._estimate_query = "SELECT TABLE_ROWS FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = '{0}';"
        self._include_id = include_id

    def _get_data_query(self, table, arrow_schema):
        cols = ','.join([x.name for x in arrow_schema])
        return  self._data_query.format(cols, table)

    def _conversions(self):
       return { }

    def _get_record_batch_schema(self, table, max_text_size, max_binary_size):
        reader = read_arrow_batches_from_odbc(
            query=self._schema_query.format(table),
            connection_string=self._connection_string,
            batch_size=1,
            max_text_size=max_text_size,
            max_binary_size=max_binary_size,
        )

        schema = reader.schema
        if not self._include_id:
            # Remove the _id column.
            return pa.schema([field for field in schema if field.name != '_id'])

        return schema

class OracleODBCDriver(object):
    def __init__(self, connection_string : str, upper_case_names : bool = False, ansi_conversion : bool = True):
        """
        Initializes the driver class.

        Parameters
        ----------
        connection_string : str
            Standard ODBC connection string used for connecting to Oracle.
            Example:
            'DSN={OracleODBC-19};Server=127.0.0.1;Port=1521;Uid=c##test;Pwd=test;DBQ=XE;'
        upper_case_names : bool
          Convert table names to uppercase similar to unqouted names behavior in Oracle queries.
          By default false.
        ansi_conversion : bool
          Convert Number(38,0) into int64s in xGT if true. Otherwise, they are stored as floats.
          This based on the ANSI int conversion Oracle does and reverses that. By default true.
        """
        self._connection_string = connection_string
        if upper_case_names:
            self._schema_query = "SELECT * FROM {0} WHERE ROWNUM <= 1"
            self._data_query = "SELECT * FROM {0}"
        else:
            self._schema_query = "SELECT * FROM \"{0}\" WHERE ROWNUM <= 1"
            self._data_query = "SELECT * FROM \"{0}\""
        self._estimate_query="SELECT NUM_ROWS FROM ALL_TABLES WHERE TABLE_NAME = '{0}'"
        self._ansi_conversion = ansi_conversion

    def _get_data_query(self, table, arrow_schema):
        return self._data_query.format(table)

    def _conversions(self):
        if self._ansi_conversion:
            return { pa.decimal128(38, 0) : pa.int64() }
        else:
            return { }

    def _get_record_batch_schema(self, table, max_text_size, max_binary_size):
        reader = read_arrow_batches_from_odbc(
            query=self._schema_query.format(table),
            connection_string=self._connection_string,
            batch_size=1,
            max_text_size=max_text_size,
            max_binary_size=max_binary_size,
        )
        return reader.schema

class SAPODBCDriver(object):
    def __init__(self, connection_string : str):
        """
        Initializes the driver class.

        Parameters
        ----------
        connection_string : str
            Standard ODBC connection string used for connecting to the ODBC applications.
            Example:
            'Driver={AES};Server=127.0.0.1;Port=3306;Database=test;Uid=test;Pwd=foo;'
        """
        self._connection_string = connection_string
        self._schema_query = "SELECT TOP 1 * FROM {0};"
        self._data_query = "SELECT * FROM {0};"
        self._estimate_query="SELECT TABLE_ROWS FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = '{0}';"

    def _get_data_query(self, table, arrow_schema):
        return  self._data_query.format(table)

    def _conversions(self):
       return { }

    def _get_record_batch_schema(self, table, max_text_size, max_binary_size):
        reader = read_arrow_batches_from_odbc(
            query=self._schema_query.format(table),
            connection_string=self._connection_string,
            batch_size=1,
            max_text_size=max_text_size,
            max_binary_size=max_binary_size,
        )
        return reader.schema

class SnowflakeODBCDriver(object):
    def __init__(self, connection_string : str, ansi_conversion : bool = True):
        """
        Initializes the driver class.

        Parameters
        ----------
        connection_string : str
            Standard ODBC connection string used for connecting to Snowflake.
            Example:
            'DSN=snowflake;Database=test;Warehouse=test;Uid=test;Pwd=test;'
        ansi_conversion : bool
          Convert Number(38,0) into int64s in xGT if true. Otherwise, they are stored as floats.
          This based on the ANSI int conversion Snowflake does and reverses that. By default true.
        """
        self._connection_string = connection_string
        self._schema_query = "SELECT * FROM {0} LIMIT 1;"
        self._data_query = "SELECT * FROM {0};"
        self._estimate_query="SELECT TABLE_ROWS FROM INFORMATION_SCHEMA.TABLES WHERE TABLE_NAME = '{0}';"
        self._ansi_conversion = ansi_conversion

    def _get_data_query(self, table, arrow_schema):
        return self._data_query.format(table)

    def _conversions(self):
        if self._ansi_conversion:
            return { pa.decimal128(38, 0) : pa.int64() }
        else:
            return { }

    def _get_record_batch_schema(self, table, max_text_size, max_binary_size):
        reader = read_arrow_batches_from_odbc(
            query=self._schema_query.format(table),
            connection_string=self._connection_string,
            batch_size=1,
            max_text_size=max_text_size,
            max_binary_size=max_binary_size,
        )
        return reader.schema

ODBCDriverTypes = Union[SQLODBCDriver, MongoODBCDriver, OracleODBCDriver,
                        SAPODBCDriver, SnowflakeODBCDriver]

class ODBCConnector(object):
    def __init__(self, xgt_server : xgt.Connection, odbc_driver : ODBCDriverTypes):
        """
        Initializes the connector class.

        Parameters
        ----------
        xgt_server : xgt.Connection
            Connection object to xGT.
        odbc_driver : SQLODBCDriver
            Connection object to ODBC.
        """
        self._xgt_server = xgt_server
        self._default_namespace = xgt_server.get_default_namespace()
        self._driver = odbc_driver

    def get_xgt_schemas(self, tables : Iter[str] = None, max_text_size : int = None,
                        max_binary_size : int = None) -> Dict:
        """
        Retrieve a dictionary containing the schema information for all of
        the tables requested and their mappings.

        Parameters
        ----------
        tables : iterable
            List of requested tables.
        max_text_size : int
            The upper limit on the buffers used when transferring ODBC variable-length text fields.
            When using VARCHAR from a database, if a limit isn't set for the length of the strings
            like VARCHAR(255), the schema size of each string entry could be whatever the max size of
            database uses for each entry when reporting to ODBC. For instance, each string in Snowflake
            has an upper limit of 16MB length. This means when allocating the buffers to store the ODBC
            batch_size would be 16MB multiplied by the batch_size. This parameter will impose a limit on
            each string length when transferring. Default is determined by the database.
        max_binary_size : int
            The upper limit on the buffers used when transferring ODBC variable-length binary fields.
            When using VARBINARY from a database, if a limit isn't set for the length of binary data
            like VARBINARY(255), the schema size of each binary entry could be whatever the max size of
            database uses for each entry when reporting to ODBC. This parameter will impose a limit on
            each binary field length when transferring. Default is determined by the database.

        Returns
        -------
        dict
            Dictionary containing the schema information of the tables,
            vertices, and edges requested.
        """
        if tables is None:
            tables = [ ]
        mapping_vertices = { }
        mapping_edges = { }
        mapping_tables = { }
        result = {'vertices' : dict(), 'edges' : dict(), 'tables' : dict()}

        for val in tables:
            self.__get_mapping(val, mapping_tables, mapping_vertices, mapping_edges)

        for table in mapping_tables:
            schema = self.__extract_xgt_table_schema(table, mapping_tables, max_text_size, max_binary_size)
            result['tables'][table] = schema

        for table in mapping_vertices:
            schema = self.__extract_xgt_table_schema(table, mapping_vertices, max_text_size, max_binary_size)
            result['vertices'][table] = schema

        for table in mapping_edges:
            schema = self.__extract_xgt_table_schema(table, mapping_edges, max_text_size, max_binary_size)
            result['edges'][table] = schema

        return result

    def create_xgt_schemas(self, xgt_schemas : Map, append : bool = False,
                           force : bool = False, easy_edges : bool = False) -> None:
        """
        Creates table, vertex and/or edge frames in Rocketgraph xGT.

        This function first infers the schemas for all of the needed frames in xGT to
        store the requested data.
        Then those frames are created in xGT.

        Parameters
        ----------
        xgt_schemas : dict
            Dictionary containing schema information for vertex and edge frames
            to create in xGT.
            This dictionary can be the value returned from the
            :py:meth:`~ODBCConnector.get_xgt_schemas` method.
        append : boolean
            Set to true when the xGT frames are already created and holding data
            that should be appended to.
            Set to false when the xGT frames are to be newly created (removing
            any existing frames with the same names prior to creation).
        force : boolean
            Set to true to force xGT to drop edges when a vertex frame has dependencies.
        easy_edges : boolean
            Set to true to create a basic vertex class with key column for any edges
            without corresponding vertex frames.

        Returns
        -------
            None
        """
        if not append:
            if easy_edges:
                for edge, schema in xgt_schemas['edges'].items():
                    src = schema['mapping']['source']
                    trg = schema['mapping']['target']
                    if src not in xgt_schemas['vertices']:
                        v_key = schema['mapping']['source_key']
                        v_type = 'int'
                        for element in schema['xgt_schema']:
                            if v_key == element[0]:
                                v_type = element[1]
                        xgt_schemas['vertices'][src] = { 'xgt_schema': [['key', v_type]], 'temp_creation' : True, 'mapping' : { 'frame' : src, 'key' : 'key' } }
                    if trg not in xgt_schemas['vertices']:
                        v_key = schema['mapping']['target_key']
                        v_type = 'int'
                        for element in schema['xgt_schema']:
                            if v_key == element[0]:
                                v_type = element[1]
                        xgt_schemas['vertices'][trg] = { 'xgt_schema': [['key', v_type]], 'temp_creation' : True, 'mapping' : { 'frame' : trg, 'key' : 'key' } }
            for _, schema in xgt_schemas['tables'].items():
                self._xgt_server.drop_frame(schema['mapping']['frame'])

            for _, schema in xgt_schemas['edges'].items():
                self._xgt_server.drop_frame(schema['mapping']['frame'])

            for _, schema in xgt_schemas['vertices'].items():
                try:
                    self._xgt_server.drop_frame(schema['mapping']['frame'])
                except xgt.XgtFrameDependencyError as e:
                    if force:
                        # Would be better if this could be done without doing this.
                        edge_frames = str(e).split(':')[-1].split(' ')[1:]
                        for edge in edge_frames:
                            self._xgt_server.drop_frame(edge)
                        self._xgt_server.drop_frame(schema['mapping']['frame'])
                    else:
                        raise e

            for table, schema in xgt_schemas['tables'].items():
                self._xgt_server.create_table_frame(name = schema['mapping']['frame'], schema = schema['xgt_schema'])

            remove_list = []
            for vertex, schema in xgt_schemas['vertices'].items():
                key = schema['mapping']['key']
                if isinstance(key, int):
                    key = schema['xgt_schema'][key][0]
                self._xgt_server.create_vertex_frame(name = schema['mapping']['frame'], schema = schema['xgt_schema'], key = key)
                if 'temp_creation' in schema:
                    remove_list.append(vertex)

            for vertex in remove_list:
                xgt_schemas['vertices'].pop(vertex)

            for edge, schema in xgt_schemas['edges'].items():
                src = schema['mapping']['source']
                trg = schema['mapping']['target']
                src_key = schema['mapping']['source_key']
                trg_key = schema['mapping']['target_key']
                if isinstance(src_key, int):
                    src_key = schema['xgt_schema'][src_key][0]
                if isinstance(trg_key, int):
                    trg_key = schema['xgt_schema'][trg_key][0]
                self._xgt_server.create_edge_frame(name = schema['mapping']['frame'], schema = schema['xgt_schema'],
                                                   source = src, target = trg, source_key = src_key, target_key = trg_key)

    def transfer_to_xgt(self, tables : Iter = None, append : bool = False, force : bool = False,
                        easy_edges : bool = False, batch_size : int = 10000, transaction_size : int = 0,
                        max_text_size : int = None, max_binary_size : int = None,
                        column_mapping : Optional[Map[str, Union[str, int]]] = None,
                        suppress_errors : bool = False, row_filter : str = None, on_duplicate_keys : str = "error") -> None:
        """
        Copies data from the ODBC application to Rocketgraph xGT.

        This function first infers the schemas for all of the needed frames in xGT to
        store the requested data.
        Then those frames are created in xGT.
        Finally, all of the tables, vertices, and all of the edges are copied,
        one frame at a time, from the ODBC application to xGT.

        Parameters
        ----------
        tables : Iterable
            List of requested tables names.
            May be a tuple specify a mapping to xGT types. See documentation: :ref:`mapping-sql-label` or `Web Docs <https://trovares.github.io/trovares_connector/odbc/index.html#mapping-sql-tables-to-graphs>`_.
        append : boolean
            Set to true when the xGT frames are already created and holding data
            that should be appended to.
            Set to false when the xGT frames are to be newly created (removing
            any existing frames with the same names prior to creation).
        force : boolean
            Set to true to force xGT to drop edges when a vertex frame has dependencies.
        easy_edges : boolean
            Set to true to create a basic vertex class wtih key column for any edges
            without corresponding vertex frames.
        batch_size : int
            Number of rows to transfer at once. Defaults to 10000.
        transaction_size : int
            Number of rows to treat as a single transaction to xGT. Defaults to 0.
            Should be a multiple of the batch size and greater than the batch size.
            0 means treat all rows as a single transaction.
        max_text_size : int
            The upper limit on the buffers used when transferring ODBC variable-length text fields.
            When using VARCHAR from a database, if a limit isn't set for the length of the strings
            like VARCHAR(255), the schema size of each string entry could be whatever the max size of
            database uses for each entry when reporting to ODBC. For instance, each string in Snowflake
            has an upper limit of 16MB length. This means when allocating the buffers to store the ODBC
            batch_size would be 16MB multiplied by the batch_size. This parameter will impose a limit on
            each string length when transferring. Default is determined by the database.
        max_binary_size : int
            The upper limit on the buffers used when transferring ODBC variable-length binary fields.
            When using VARBINARY from a database, if a limit isn't set for the length of binary data
            like VARBINARY(255), the schema size of each binary entry could be whatever the max size of
            database uses for each entry when reporting to ODBC. This parameter will impose a limit on
            each binary field length when transferring. Default is determined by the database.
        column_mapping : dictionary
            Maps the frame column names to SQL columns for the ingest. The key of
            each element is a frame column name. The value is either the name of the
            SQL column (from the table) or the table column index.
        suppress_errors : bool
            If true, will continue to insert data if an ingest error is encountered,
            placing the first 1000 errors in the job history. If false, stops on
            first error and raises.  Defaults to False.
        row_filter : str
            TQL fragment used to filter, modify and parameterize the raw data from
            the input to produce the row data fed to the frame.
        on_duplicate_keys : {‘error’, ‘skip’, 'skip_same'}, default 'error'
            Specifies what to do upon encountering a duplicate vertex key.
            Only works for vertex frames. Is ignored for table and edge frames.
            Allowed values are :
            - 'error', raise an Exception when a duplicate key is found.
            - 'skip', skip duplicate keys without raising.
            - 'skip_same', skip duplicate keys if the row is exactly the same without raising.

        Returns
        -------
            None
        """
        if transaction_size > 0 and (transaction_size < batch_size or transaction_size % batch_size != 0):
            raise ValueError("Transaction size needs to be a multiple of the batch size and >= the batch size of " + str(batch_size))
        xgt_schema = self.get_xgt_schemas(tables, max_text_size, max_binary_size)
        self.create_xgt_schemas(xgt_schema, append, force, easy_edges)
        self.copy_data_to_xgt(xgt_schema, batch_size, transaction_size,
                              max_text_size, max_binary_size, column_mapping,
                              suppress_errors, row_filter, on_duplicate_keys)

    def transfer_query_to_xgt(self, query : str = None, mapping : Union[Map, tuple] = None, append : bool = False,
                              force : bool = False, easy_edges : bool = False, batch_size : int = 10000,
                              transaction_size : int = 0, max_text_size : int = None,
                              max_binary_size : int = None, column_mapping : Optional[Map[str, Union[str, int]]] = None,
                              suppress_errors : bool = False,
                              row_filter : str = None, on_duplicate_keys : str = "error") -> None:
        """
        Copies data from the ODBC application to Rocketgraph xGT.

        This function first infers the schemas for the query.
        Then it maps to the type specificed in mapping.
        Finally, the data is copied from the ODBC application to xGT.

        Parameters
        ----------
        query : string
            SQL query to execute and insert into xGT. Syntax depends on the SQL syntax of the database you are connecting to.
        mapping :
            May be a tuple specify a mapping to xGT types. See documentation: :ref:`mapping-sql-label` or `Web Docs <https://trovares.github.io/trovares_connector/odbc/index.html#mapping-sql-tables-to-graphs>`_.
        append : boolean
            Set to true when the xGT frames are already created and holding data
            that should be appended to.
            Set to false when the xGT frames are to be newly created (removing
            any existing frames with the same names prior to creation).
        force : boolean
            Set to true to force xGT to drop edges when a vertex frame has dependencies.
        easy_edges : boolean
            Set to true to create a basic vertex class with key column for any edges
            without corresponding vertex frames.
        batch_size : int
            Number of rows to transfer at once. Defaults to 10000.
        transaction_size : int
            Number of rows to treat as a single transaction to xGT. Defaults to 0.
            Should be a multiple of the batch size and greater than the batch size.
            0 means treat all rows as a single transaction.
        max_text_size : int
            The upper limit on the buffers used when transferring ODBC variable-length text fields.
            When using VARCHAR from a database, if a limit isn't set for the length of the strings
            like VARCHAR(255), the schema size of each string entry could be whatever the max size of
            database uses for each entry when reporting to ODBC. For instance, each string in Snowflake
            has an upper limit of 16MB length. This means when allocating the buffers to store the ODBC
            batch_size would be 16MB multiplied by the batch_size. This parameter will impose a limit on
            each string length when transferring. Default is determined by the database.
        max_binary_size : int
            The upper limit on the buffers used when transferring ODBC variable-length binary fields.
            When using VARBINARY from a database, if a limit isn't set for the length of binary data
            like VARBINARY(255), the schema size of each binary entry could be whatever the max size of
            database uses for each entry when reporting to ODBC. This parameter will impose a limit on
            each binary field length when transferring. Default is determined by the database.
        column_mapping : dictionary
            Maps the frame column names to SQL columns for the ingest. The key of
            each element is a frame column name. The value is either the name of the
            SQL column (from the table) or the table column index.
        suppress_errors : bool
            If true, will continue to insert data if an ingest error is encountered,
            placing the first 1000 errors in the job history. If false, stops on
            first error and raises.  Defaults to False.
        row_filter : str
            TQL fragment used to filter, modify and parameterize the raw data from
            the input to produce the row data fed to the frame.
        on_duplicate_keys : {‘error’, ‘skip’, 'skip_same'}, default 'error'
            Specifies what to do upon encountering a duplicate vertex key.
            Only works for vertex frames. Is ignored for table and edge frames.
            Allowed values are :
            - 'error', raise an Exception when a duplicate key is found.
            - 'skip', skip duplicate keys without raising.
            - 'skip_same', skip duplicate keys if the row is exactly the same without raising.

        Returns
        -------
            array of transfer information in the form of [row count, byte count]
        """
        if transaction_size > 0 and (transaction_size < batch_size or transaction_size % batch_size != 0):
            raise ValueError("Transaction size needs to be a multiple of batch size and >= the batch size of " + str(batch_size))
        return self.__copy_query_data_to_xgt(query, mapping, append, force, easy_edges,
                                             batch_size, transaction_size, max_text_size, max_binary_size,
                                             column_mapping, suppress_errors, row_filter, on_duplicate_keys)

    def copy_data_to_xgt(self, xgt_schemas : Map, batch_size : int = 10000, transaction_size : int = 0,
                         max_text_size : int = None, max_binary_size : int = None,
                         column_mapping : Optional[Map[str, Union[str, int]]] = None,
                         suppress_errors : bool = False, row_filter : str = None,
                         on_duplicate_keys : str = "error") -> None:
        """
        Copies data from the ODBC application to the requested table, vertex and/or edge frames
        in Rocketgraph xGT.

        This function copies data from the ODBC application to xGT for all of the tables, vertices
        and edges, one frame at a time.

        Parameters
        ----------
        xgt_schemas : dict
            Dictionary containing schema information for table, vertex and edge frames
            to create in xGT.
            This dictionary can be the value returned from the
            :py:meth:`~ODBCConnector.get_xgt_schemas` method.
        batch_size : int
            Number of rows to transfer at once. Defaults to 10000.
        transaction_size : int
            Number of rows to treat as a single transaction to xGT. Defaults to 0.
            Should be a multiple of the batch size and greater than the batch size.
            0 means treat all rows as a single transaction.
        max_text_size : int
            The upper limit on the buffers used when transferring ODBC variable-length text fields.
            When using VARCHAR from a database, if a limit isn't set for the length of the strings
            like VARCHAR(255), the schema size of each string entry could be whatever the max size of
            database uses for each entry when reporting to ODBC. For instance, each string in Snowflake
            has an upper limit of 16MB length. This means when allocating the buffers to store the ODBC
            batch_size would be 16MB multiplied by the batch_size. This parameter will impose a limit on
            each string length when transferring. Default is determined by the database.
        max_binary_size : int
            The upper limit on the buffers used when transferring ODBC variable-length binary fields.
            When using VARBINARY from a database, if a limit isn't set for the length of binary data
            like VARBINARY(255), the schema size of each binary entry could be whatever the max size of
            database uses for each entry when reporting to ODBC. This parameter will impose a limit on
            each binary field length when transferring. Default is determined by the database.
        column_mapping : dictionary
            Maps the frame column names to SQL columns for the ingest. The key of
            each element is a frame column name. The value is either the name of the
            SQL column (from the table) or the table column index.
        suppress_errors : bool
            If true, will continue to insert data if an ingest error is encountered,
            placing the first 1000 errors in the job history. If false, stops on
            first error and raises.  Defaults to False.
        row_filter : str
            TQL fragment used to filter, modify and parameterize the raw data from
            the input to produce the row data fed to the frame.
        on_duplicate_keys : {‘error’, ‘skip’, 'skip_same'}, default 'error'
            Specifies what to do upon encountering a duplicate vertex key.
            Only works for vertex frames. Is ignored for table and edge frames.
            Allowed values are :
            - 'error', raise an Exception when a duplicate key is found.
            - 'skip', skip duplicate keys without raising.
            - 'skip_same', skip duplicate keys if the row is exactly the same without raising.

        Returns
        -------
            None
        """
        estimate = 0
        def estimate_size(table):
            estimate = 0
            reader = read_arrow_batches_from_odbc(
                query=self._driver._estimate_query.format(table),
                connection_string=self._driver._connection_string,
                batch_size=batch_size,
                max_text_size=max_text_size,
                max_binary_size=max_binary_size,
            )
            for batch in reader:
                for _, row in batch.to_pydict().items():
                    for item in row:
                        if isinstance(item, int):
                            estimate += item
            return estimate
        try:
            for table, schema in xgt_schemas['tables'].items():
                estimate += estimate_size(table)
            for table, schema in xgt_schemas['vertices'].items():
                estimate += estimate_size(table)
            for table, schema in xgt_schemas['edges'].items():
                estimate += estimate_size(table)
        except Exception as e:
            pass

        with ProgressDisplay(estimate) as progress_bar:
            for table, schema in xgt_schemas['tables'].items():
                self.__copy_data(self._driver._get_data_query(
                    table, schema['arrow_schema']), schema['mapping']['frame'],
                    schema['arrow_schema'], progress_bar, batch_size,
                    transaction_size, max_text_size, max_binary_size,
                    column_mapping, suppress_errors, row_filter,
                    on_duplicate_keys)
            for table, schema in xgt_schemas['vertices'].items():
                self.__copy_data(self._driver._get_data_query(
                    table, schema['arrow_schema']), schema['mapping']['frame'],
                    schema['arrow_schema'], progress_bar, batch_size,
                    transaction_size, max_text_size, max_binary_size,
                    column_mapping, suppress_errors, row_filter,
                    on_duplicate_keys)
            for table, schema in xgt_schemas['edges'].items():
                self.__copy_data(self._driver._get_data_query(
                    table, schema['arrow_schema']), schema['mapping']['frame'],
                    schema['arrow_schema'], progress_bar, batch_size,
                    transaction_size, max_text_size, max_binary_size,
                    column_mapping, suppress_errors, row_filter,
                    on_duplicate_keys)

    def transfer_to_odbc(self, vertices : Iter[str] = None,
                         edges : Iter[str] = None,
                         tables : Iter[str] = None, namespace : str = None,
                         batch_size : int = 10000) -> None:
        """
        Copies data from Rocketgraph xGT to an ODBC application.

        Parameters
        ----------
        vertices : iterable
            List of requested vertex frame names.
            May be a tuple specifying: (xgt_frame_name, database_table_name).
        edges : iterable
            List of requested edge frame names.
            May be a tuple specifying: (xgt_frame_name, database_table_name).
        tables : iterable
            List of requested table frame names.
            May be a tuple specifying: (xgt_frame_name, database_table_name).
        namespace : str
            Namespace for the selected frames.
            If none will use the default namespace.
        batch_size : int
            Number of rows to transfer at once. Defaults to 10000.

        Returns
        -------
            None
        """
        if isinstance(self._driver, OracleODBCDriver):
            raise XgtNotImplementedError("Oracle not supported for transferring to.")
        xgt_server = self._xgt_server
        if namespace == None:
            namespace = self._default_namespace
        if vertices == None and edges == None and tables == None:
            vertices = [(frame.name, frame.name) for frame in xgt_server.get_frames(namespace=namespace, frame_type='vertex')]
            edges = [(frame.name, frame.name) for frame in xgt_server.get_frames(namespace=namespace, frame_type='edge')]
            tables = [(frame.name, frame.name) for frame in xgt_server.get_frames(namespace=namespace, frame_type='table')]
            namespace = None
        if vertices == None:
            vertices = []
        if edges == None:
            edges = []
        if tables == None:
            tables = []

        final_vertices = []
        final_edges = []
        final_tables = []

        for vertex in vertices:
            if isinstance(vertex, str):
                final_vertices.append((vertex, vertex))
            else:
                final_vertices.append(vertex)
        for edge in edges:
            if isinstance(edge, str):
                final_edges.append((edge, edge))
            else:
                final_edges.append(edge)
        for table in tables:
            if isinstance(table, str):
                final_tables.append((table, table))
            else:
                final_tables.append(table)

        estimate = 0

        for vertex in final_vertices:
            estimate += xgt_server.get_frame(vertex[0]).num_rows
        for edge in final_edges:
            estimate += xgt_server.get_frame(edge[0]).num_rows
        for table in final_tables:
            estimate += xgt_server.get_frame(table[0]).num_rows

        with ProgressDisplay(estimate) as progress_bar:
            for table in final_vertices + final_edges + final_tables:
                frame, table = table
                reader = self.__arrow_reader(frame)
                batch_reader = reader.to_reader()

                _, target_schema = self.__get_xgt_schema(table)
                schema = reader.schema
                final_schema = [xgt_field.with_name(database_field.name) for database_field, xgt_field in zip(target_schema, schema)]
                final_schema = pa.schema(final_schema)
                schema = final_schema
                final_names = [database_field.name for database_field in target_schema]
                def iter_record_batches():
                    for batch in batch_reader:
                        table = pa.Table.from_pandas(batch.to_pandas(integer_object_nulls=True, date_as_object=True, timestamp_as_object=True))
                        table = table.rename_columns(final_names).to_batches()
                        for batch in table:
                            yield batch
                            progress_bar.show_progress(batch.num_rows)

                final_reader = pa.ipc.RecordBatchReader.from_batches(schema, iter_record_batches())
                insert_into_table(
                    connection_string=self._driver._connection_string,
                    chunk_size=batch_size,
                    table=table,
                    reader=final_reader,
                )

    def __build_flight_path(self, frame_name, column_mapping = None,
                            suppress_errors = False, row_filter = None,
                            on_duplicate_keys = 'error'):
        if '__' in frame_name:
            # Split by '__' and use the first part as the namespace
            namespace, name = frame_name.split('__', 1)
        else:
            # Use the default namespace if no '__' is found
            namespace, name = self._default_namespace, frame_name

        path = (namespace, name)

        self.__validate_column_mapping(column_mapping)
        if row_filter is None and column_mapping is not None:
            map_values = ".map_column_names=[" + \
                ','.join(f"{key}:{value}" for key, value in column_mapping.items()
                    if isinstance(value, str)) + "]"
            path += (map_values,)
            map_values = ".map_column_ids=[" + \
                ','.join(f"{key}:{value}" for key, value in column_mapping.items()
                    if isinstance(value, int)) + "]"
            path += (map_values,)

        suppress_errors_option = ".suppress_errors=" + str(suppress_errors).lower()
        on_duplicate_keys_option = ".on_duplicate_keys=" + str(on_duplicate_keys).lower()
        path += (suppress_errors_option,)
        path += (on_duplicate_keys_option,)

        if row_filter is not None:
            row_filter_value = f'.row_filter="{row_filter}"'
            path += (row_filter_value,)

        return path

    def __arrow_writer(self, frame_name, schema, column_mapping, suppress_errors, row_filter, on_duplicate_keys):
        arrow_conn = self._xgt_server.arrow_conn
        flight_path = self.__build_flight_path(frame_name, column_mapping, suppress_errors, row_filter, on_duplicate_keys)
        writer, metadata = arrow_conn.do_put(
            pf.FlightDescriptor.for_path(*flight_path),
            schema)
        return (writer, metadata)

    def __arrow_reader(self, frame_name):
        arrow_conn = self._xgt_server.arrow_conn
        return arrow_conn.do_get(pf.Ticket(self._default_namespace + '__' + frame_name))

    def __copy_data(self, query_for_extract, frame, schema, progress_bar, batch_size,
                    transaction_size, max_text_size, max_binary_size, column_mapping,
                    suppress_errors, row_filter, on_duplicate_keys):
        try:
            reader = read_arrow_batches_from_odbc(
                query=query_for_extract,
                connection_string=self._driver._connection_string,
                batch_size=batch_size,
                max_text_size=max_text_size,
                max_binary_size=max_binary_size,
            )
            count = 0
            writer, metadata = self.__arrow_writer(frame, schema, column_mapping, suppress_errors, row_filter, on_duplicate_keys)
            for batch in reader:
                # Process arrow batches
                writer.write(batch)
                progress_bar.show_progress(batch.num_rows)
                count += batch.num_rows
                # Start a new transaction
                if transaction_size > 0 and count >= transaction_size:
                    count = 0
                    if (suppress_errors):
                        self.__check_for_error(frame, schema, writer, metadata)
                    writer.close()
                    writer, metadata = self.__arrow_writer(frame, schema, column_mapping, suppress_errors, row_filter, on_duplicate_keys)

            if (suppress_errors):
                self.__check_for_error(frame, schema, writer, metadata)

            writer.close()
        except arrow_odbc.error.Error as e:
            if "too large to be written" in str(e):
                raise RuntimeError(
                    f"{e}\nHint: Try building the query yourself with transfer_query_to_xgt and casting oversized columns with CAST(column AS VARCHAR(n)) "
                    "or changing the schema of the table to use a fixed size VARCHAR(n). "
                    "Variable text width columns sometimes report incorrect maximum sizes from drivers. "
                ) from e
            raise

    def __check_for_error(self, frame, schema, writer, metadata):
        # Write an empty batch with metadata to indicate we are done.
        empty = [[]] * len(schema)
        empty_batch = pa.RecordBatch.from_arrays(empty, schema = schema)
        metadata_end = struct.pack('<i', 0)
        writer.write_with_metadata(empty_batch, metadata_end)
        buf = metadata.read()
        job_proto = sch_proto.JobStatus()
        if buf is not None:
          job_proto.ParseFromString(buf.to_pybytes())

        job = xgt.Job(self._xgt_server, job_proto)
        job_data = job.get_ingest_errors()

        if job_data is not None and len(job_data) > 0:
          raise xgt.XgtIOError(self.__create_ingest_error_message(frame, job), job = job)

    def __create_ingest_error_message(self, name, job):
        num_errors = job.total_ingest_errors

        error_string = ('Errors occurred when inserting data into frame '
                        f'{name}.\n')

        error_string += f'  {num_errors} line'
        if num_errors > 1:
          error_string += 's'

        error_string += (' had insertion errors.\n'
                         '  Lines without errors were inserted into the frame.\n'
                         '  To see the number of rows in the frame, run "'
                         f'{name}.num_rows".\n'
                         '  To see the data in the frame, run "'
                         f'{name}.get_data()".\n'
                         '  To see additional errors:\n'
                         '    try:\n'
                         '      foo.transfer_to_xgt(...)\n'
                         '    except xgt.XgtIOError as e:\n'
                         '      error_rows = e.job.get_ingest_errors()\n'
                         '      print(error_rows)\n')

        extra_text = ''
        if num_errors > 10:
          extra_text = ' first 10'

        error_string += (f'Errors associated with the{extra_text} lines '
                         'that could not be inserted are shown below:')

        # Only print the first 10 messages.
        for error in job.get_ingest_errors(0, 10):
          delim = ','

          # Will process a list of strings. Convert to this format.
          if isinstance(error, str):
            error_cols = error.split(delim)
          elif isinstance(error, list):
            error_cols = [str(elem) for elem in error]
          else:
            raise xgt.XgtIOError("Error processing ingest error message.")

          # The first comma separated fields of the error string are the error
          # description, file name, and line number.
          error_explanation = "" if len(error_cols) < 1 else error_cols[0]
          error_file_name = "odbc_transfer"
          error_line_number = "" if len(error_cols) < 3 else error_cols[2]

          # The second part of the error string contains the line that caused the
          # error. The line contains comma separated fields so we need to re-join
          # these comma separated portions to get back the line.
          line_with_error = \
              "" if len(error_cols) < 4 else delim.join(error_cols[3:])

          error_string += f"\n {error_explanation}"

        return error_string

    def __get_xgt_schema(self, table, max_text_size = None, max_binary_size = None):
        schema = self._driver._get_record_batch_schema(table, max_text_size, max_binary_size)
        return (_infer_xgt_schema_from_pyarrow_schema(schema, self._driver._conversions()), schema)

    def __extract_xgt_table_schema(self, table, mapping, max_text_size, max_binary_size):
        xgt_schema, arrow_schema = self.__get_xgt_schema(table, max_text_size, max_binary_size)
        return {'xgt_schema' : xgt_schema, 'arrow_schema' : arrow_schema, 'mapping' : mapping[table]}

    def __get_mapping(self, val, mapping_tables, mapping_vertices, mapping_edges):
        if isinstance(val, str):
            mapping_tables[val] = {'frame' : val}
        elif isinstance(val, tuple):
            # ('table', X, ...)
            if isinstance(val[1], str):
                # ('table', (...), ...)
                if len(val) == 3:
                    # ('table', (), ...)
                    if len(val[2]) == 1:
                        mapping_vertices[val[0]] = {'frame' : val[1], 'key' : val[2][0]}
                    # ('table', (,), ...)
                    elif len(val[2]) == 4:
                        mapping_edges[val[0]] = {'frame' : val[1], 'source' : val[2][0],
                                                 'target' : val[2][1], 'source_key' : val[2][2],
                                                 'target_key' : val[2][3]}
                else:
                    mapping_tables[val[0]] = {'frame' : val[1]}
            elif isinstance(val[1], tuple) and len(val[1]) == 1:
                mapping_vertices[val[0]] = {'frame' : val[0], 'key' : val[1][0]}
            elif isinstance(val[1], tuple) and len(val[1]) == 4:
                mapping_edges[val[0]] = {'frame' : val[0], 'source' : val[1][0],
                                         'target' : val[1][1], 'source_key' : val[1][2],
                                         'target_key' : val[1][3]}
            elif isinstance(val[1], dict):
                if len(val[1]) == 1:
                    mapping_tables[val[0]] = val[1]
                elif len(val[1]) == 2:
                    mapping_vertices[val[0]] = val[1]
                elif len(val[1]) == 5:
                    mapping_edges[val[0]] = val[1]
                else:
                    raise ValueError("Dictionary format incorrect for " + str(val[0]))
            else:
                raise ValueError("Argument format incorrect for " + str(val))

    def __copy_query_data_to_xgt(self, query, mapping, append, force, easy_edges,
                                 batch_size, transaction_size, max_text_size, max_binary_size,
                                 column_mapping, suppress_errors, row_filter, on_duplicate_keys):
        try:
            estimate = 0
            mapping_vertices = { }
            mapping_edges = { }
            mapping_tables = { }
            result = {'vertices' : dict(), 'edges' : dict(), 'tables' : dict()}
            self.__get_mapping(mapping, mapping_tables, mapping_vertices, mapping_edges)

            with ProgressDisplay(estimate) as progress_bar:
                reader = read_arrow_batches_from_odbc(
                    query=query,
                    connection_string=self._driver._connection_string,
                    batch_size=batch_size,
                    max_text_size=max_text_size,
                    max_binary_size=max_binary_size,
                )
                arrow_schema = reader.schema
                xgt_schema = _infer_xgt_schema_from_pyarrow_schema(arrow_schema, self._driver._conversions())
                for table in mapping_tables:
                    schema = {'xgt_schema' : xgt_schema, 'arrow_schema' : arrow_schema, 'mapping' : mapping_tables[table]}
                    result['tables'][table] = schema
                    frame = schema['mapping']['frame']

                for table in mapping_vertices:
                    schema = {'xgt_schema' : xgt_schema, 'arrow_schema' : arrow_schema, 'mapping' : mapping_vertices[table]}
                    result['vertices'][table] = schema
                    frame = schema['mapping']['frame']

                for table in mapping_edges:
                    schema = {'xgt_schema' : xgt_schema, 'arrow_schema' : arrow_schema, 'mapping' : mapping_edges[table]}
                    result['edges'][table] = schema
                    frame = schema['mapping']['frame']

                self.create_xgt_schemas(result, append, force, easy_edges)
                writer, metadata = self.__arrow_writer(frame, arrow_schema, column_mapping, suppress_errors, row_filter, on_duplicate_keys)
                count = 0
                bytes_transferred = 0
                row_count = 0
                for batch in reader:
                    bytes_transferred += sum(column.nbytes for column in batch)
                    # Process arrow batches
                    writer.write(batch)
                    progress_bar.show_progress(batch.num_rows)
                    count += batch.num_rows
                    row_count += batch.num_rows
                    # Start a new transaction
                    if transaction_size > 0 and count >= transaction_size:
                        count = 0
                        if (suppress_errors):
                            self.__check_for_error(frame, arrow_schema, writer, metadata)
                        writer.close()
                        writer, metadata = self.__arrow_writer(frame, arrow_schema, column_mapping, suppress_errors, row_filter, on_duplicate_keys)

                if (suppress_errors):
                    self.__check_for_error(frame, arrow_schema, writer, metadata)
                writer.close()

                return row_count, bytes_transferred
        except arrow_odbc.error.Error as e:
            if "too large to be written" in str(e):
                raise RuntimeError(
                    f"{e}\nHint: Try casting oversized columns with CAST(column AS VARCHAR(n)) "
                    "or changing the schema of the table to use a fixed size VARCHAR(n). "
                    "Variable text width columns sometimes report incorrect maximum sizes from drivers. "
                ) from e
            raise

    def __validate_column_mapping(self, column_mapping):
        error_msg = ('The data type of "column_mapping" is incorrect. '
                     'Expects a dictionary with string keys and string '
                     'or integer values.')
        if column_mapping is not None:
            if not isinstance(column_mapping, Mapping):
                raise TypeError(error_msg)
            for frame_col, file_col in column_mapping.items():
                if not isinstance(file_col, (str, int)):
                    raise TypeError(error_msg)
                if not isinstance(frame_col, str):
                    raise TypeError(error_msg)
